package io.github.lonamiwebs.overgram.parser;

import io.github.lonamiwebs.overgram.utils.Utils;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

public class TLArg {
    public final String name;
    public final List<String> types;
    public final boolean flags;
    public final Flag flag;
    public final boolean genericDefinition;
    public final TLArg genericOn;

    private TLArg(final String name, final List<String> types, final boolean flags, final Flag flag,
                  final boolean genericDefinition, final TLArg genericOn) {
        this.name = name;
        this.types = types;
        this.flags = flags;
        this.flag = flag;
        this.genericDefinition = genericDefinition;
        this.genericOn = genericOn;
    }

    public static TLArg fromString(final String string, final Map<String, TLArg> genericDefinitions) {
        final String[] tmp = string.split(":");

        boolean genericDefinition = false;
        if (tmp[1].equals("Type}")) {
            tmp[1] = tmp[1].substring(0, tmp[1].length() - 1);
            tmp[0] = tmp[0].substring(1);
            genericDefinition = true;
        }

        boolean flags = false;
        if (tmp[1].equals("#")) {
            tmp[1] = "";
            flags = true;
        }

        Flag flag = null;
        final int flagIndex = tmp[1].indexOf('?');
        if (flagIndex != -1) {
            flag = Flag.fromString(tmp[1].substring(0, flagIndex));
            tmp[1] = tmp[1].substring(flagIndex + 1);
        }

        final List<String> types = new ArrayList<>(2);
        int angleBracket;
        while ((angleBracket = tmp[1].indexOf("<")) != -1) {
            types.add(tmp[1].substring(0, angleBracket));
            tmp[1] = tmp[1].substring(angleBracket + 1, tmp[1].length() - 1);
        }
        types.add(tmp[1]);

        final TLArg genericOn;
        if (!genericDefinition && types.size() == 1) {
            genericOn = genericDefinitions.get(types.get(0));
        } else {
            genericOn = null;
        }

        return new TLArg(tmp[0], types, flags, flag, genericDefinition, genericOn);
    }

    public static TLArg fromUnnamedString(final String string, final Map<String, TLArg> genericDefinitions) {
        return fromString(":" + string, genericDefinitions);
    }

    public String javaType() {
        return javaType(false);
    }

    public String javaType(boolean nested) {
        if (genericOn != null) {
            return genericOn.javaType();
        }

        if (genericDefinition) {
            if (types.size() == 1 && types.get(0).equals("Type")) {
                return "TLObject";
            } else {
                throw new UnsupportedOperationException();
            }
        }

        final StringBuilder builder = new StringBuilder();
        for (final String type : types) {
            if (type.startsWith("!")) {
                if (nested) {
                    throw new UnsupportedOperationException();
                }
                return "TLObject";
            }

            switch (type) {
                case "vector":
                case "Vector":
                    builder.append("List");
                    break;
                case "int":
                    if (flag == null && !nested) {
                        builder.append("int");
                    } else {
                        builder.append("Integer");
                    }
                    break;
                case "long":
                    if (flag == null && !nested) {
                        builder.append("long");
                    } else {
                        builder.append("Long");
                    }
                    break;
                case "int128":
                case "int256":
                    builder.append("BigInteger");
                    break;
                case "double":
                    if (flag == null && !nested) {
                        builder.append("double");
                    } else {
                        builder.append("Double");
                    }
                    break;
                case "Bool":
                    if (flag == null && !nested) {
                        builder.append("boolean");
                    } else {
                        builder.append("Boolean");
                    }
                    break;
                case "true":
                    builder.append("boolean");
                    break;
                case "string":
                    builder.append("String");
                    break;
                case "bytes":
                    builder.append("byte[]");
                    break;
                default:
                    builder.append("Abstract.");
                    builder.append(Utils.toCamelCase(type, true));
                    break;
            }
            builder.append('<');
            nested = true;
        }

        builder.setLength(builder.length() - 1);
        for (int i = 1; i < types.size(); ++i) {
            builder.append('>');
        }

        return builder.toString();
    }

    public int bigIntSize() {
        switch (types.get(types.size() - 1)) {
            case "int128":
                return 16;
            case "int256":
                return 32;
            default:
                return 0;
        }
    }

    public String javaSetCheck() {
        return types.size() == 1 && types.get(0).equals("true") ? "" : " != null";
    }

    public String javaName() {
        switch (name) {
            case "lat":
                return "latitude"; // not reserved but consistent
            case "long":
                return "longitude";
            case "static":
                return "isStatic";
            case "public":
                return "isPublic";
            case "private":
                return "isPrivate";
            case "final":
                return "last";
            default:
                return Utils.toCamelCase(name, false);
        }
    }

    @Override
    public String toString() {
        final StringBuilder builder = new StringBuilder();
        if (genericDefinition) {
            builder.append('{');
        }
        builder.append(name);
        builder.append(':');
        if (flag != null) {
            builder.append(flag.toString());
        }
        if (flags) {
            builder.append('#');
        } else {
            for (final String type : types) {
                builder.append(type);
                builder.append('<');
            }
            builder.setLength(builder.length() - 1);
            for (int i = 1; i < types.size(); ++i) {
                builder.append('>');
            }
        }
        if (genericDefinition) {
            builder.append('}');
        }

        return builder.toString();
    }

    public static final class Flag {
        public final String flagName;
        public final int flagIndex;

        private Flag(final String flagName, final int flagIndex) {
            this.flagName = flagName;
            this.flagIndex = flagIndex;
        }

        private static Flag fromString(final String string) {
            final String[] tmp = string.split("\\.");
            return new Flag(tmp[0], Integer.parseInt(tmp[1]));
        }

        @Override
        public String toString() {
            return flagName + "." + flagIndex + "?";
        }
    }
}
